from __future__ import annotations

import time
import warnings
from collections import defaultdict
from datetime import date, datetime
from itertools import chain
from pathlib import Path

import numpy as np
from construct.core import StreamError

from oct_converter.image_types import FundusImageWithMetaData, OCTVolumeWithMetaData
from oct_converter.readers.binary_structs import e2e_binary


class E2E(object):
    """Class for extracting data from Heidelberg's .e2e file format.

    Notes:
        Mostly based on description of .e2e file format here:
        https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format.

    Attributes:
        filepath: path to .e2e file for reading.
    """

    def __init__(self, filepath: str | Path) -> None:
        self.filepath = Path(filepath)
        if not self.filepath.exists():
            raise FileNotFoundError(self.filepath)

        self.power = pow(2, 10)
        self.sex = None
        self.first_name = None
        self.surname = None
        self.acquisition_date = None
        self.birthdate = None
        self.pixel_spacing = None
        self.patient_id = None

        # get initial directory structure
        with open(self.filepath, "rb") as f:
            raw = f.read(21)
            if raw == b"E2EMultipleVolumeFile":
                self.byte_skip = 64
            else:
                self.byte_skip = 0
            f.seek(self.byte_skip)
            raw = f.read(36)

            header = e2e_binary.header_structure.parse(raw)
            raw = f.read(52)
            main_directory = e2e_binary.main_directory_structure.parse(raw)

            # traverse list of main directories in first pass
            self.directory_stack = []

            current = main_directory.current
            while current != 0:
                self.directory_stack.append(current)
                f.seek(current + self.byte_skip)
                raw = f.read(52)
                directory_chunk = e2e_binary.main_directory_structure.parse(raw)
                current = directory_chunk.prev

    def read_oct_volume(
        self,
        legacy_intensity_transform: bool = False,
        scalex: float = 0.01,
        slice_thickness: float = 0.05,
    ) -> list[OCTVolumeWithMetaData]:
        """Reads OCT data.

        Args:
             legacy_intensity_transform: if True, use intensity transform used in v<=0.5.7. Defaults to False.
             scalex: Manually set scale of x axis
             slice_thickness: Manually set scale of z axis

        Returns:
            A list of OCTVolumeWithMetaData.
        """

        def _make_lut():
            LUT = []
            for i in range(0, pow(2, 16)):
                LUT.append(self.uint16_to_ufloat16(i))
            return np.array(LUT)

        LUT = _make_lut()

        with open(self.filepath, "rb") as f:
            # get all subdirectories
            chunk_stack = []
            volume_dict = {}
            for position in self.directory_stack:
                f.seek(position + self.byte_skip)
                raw = f.read(52)
                directory_chunk = e2e_binary.main_directory_structure.parse(raw)

                for ii in range(directory_chunk.num_entries):
                    raw = f.read(44)
                    chunk = e2e_binary.sub_directory_structure.parse(raw)
                    volume_string = "{}_{}_{}".format(
                        chunk.patient_db_id, chunk.study_id, chunk.series_id
                    )
                    if volume_string not in volume_dict.keys():
                        volume_dict[volume_string] = chunk.slice_id / 2
                    elif chunk.slice_id / 2 > volume_dict[volume_string]:
                        volume_dict[volume_string] = chunk.slice_id / 2

                    if chunk.start > chunk.pos:
                        chunk_stack.append([chunk.start, chunk.size])

            # initalise dict to hold all the image volumes
            volume_array_dict = {}
            volume_array_dict_additional = (
                {}
            )  # for storage of slices not caught by extraction
            laterality_dict = {}
            laterality = None
            for volume, num_slices in volume_dict.items():
                if num_slices > 0:
                    # num_slices + 1 here due to evidence that a slice was being missed off the end in extraction
                    volume_array_dict[volume] = [0] * int(num_slices + 1)

            contour_dict = defaultdict(lambda: defaultdict(dict))

            # traverse all chunks and extract slices
            for start, pos in chunk_stack:
                f.seek(start + self.byte_skip)
                raw = f.read(60)
                try:
                    # Heidelberg's updated anonymization seems to cause problems with
                    # some chunks. Observed problems include an empty raw and problems
                    # with undecodable bytes. For now, these chunks are skipped...
                    chunk = e2e_binary.chunk_structure.parse(raw)
                except Exception:
                    continue

                if chunk.type == 9:  # patient data
                    raw = f.read(127)
                    try:
                        patient_data = e2e_binary.patient_id_structure.parse(raw)
                        self.sex = patient_data.sex
                        self.first_name = patient_data.first_name
                        self.surname = patient_data.surname
                        self.patient_id = patient_data.patient_id
                        if len(str(patient_data.birthdate)) == 8:
                            # Encountered a file where birthdate had been stored as YYYYMMDD,
                            # this is an attempt to catch that.
                            self.birthdate = str(patient_data.birthdate)
                        else:
                            try:
                                julian_birthdate = (
                                    patient_data.birthdate / 64
                                ) - 14558805
                                self.birthdate = self.julian_to_ymd(julian_birthdate)
                                # TODO: There are conflicting ideas of how to parse E2E's birthdate
                                # https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format suggests the above,
                                # whereas https://github.com/neurodial/LibE2E/blob/master/E2E/dataelements/patientdataelement.cpp
                                # suggests that DOB is given as a Windows date. Neither option seems accurate to
                                # test files with known-correct birthdates. More investigation is needed.
                            except ValueError:
                                # If the julian_to_ymd function cannot parse it into a date obj,
                                # it throws a ValueError
                                self.birthdate = None
                    except Exception:
                        pass

                elif chunk.type == 10004:  # bscan metadata
                    raw = f.read(104)
                    bscan_metadata = e2e_binary.bscan_metadata.parse(raw)

                    windowsTicks = bscan_metadata.acquisitionTime
                    windowsTicksToUnixFactor = 10000000
                    secToUnixEpechFromWindowsTicks = 11644473600
                    unixtime = (
                        windowsTicks / windowsTicksToUnixFactor
                        - secToUnixEpechFromWindowsTicks
                    )
                    utc_time = time.gmtime(unixtime)
                    dt_time = datetime.fromtimestamp(time.mktime(utc_time))
                    if self.acquisition_date is None:
                        self.acquisition_date = dt_time
                    if self.pixel_spacing is None:
                        # scaley found, x and z not yet found in file
                        self.pixel_spacing = [
                            scalex,
                            bscan_metadata.scaley,
                            slice_thickness,
                        ]

                elif chunk.type == 3:  # scan preamble data
                    raw = f.read(chunk.size)
                    try:
                        pre_data = e2e_binary.pre_data.parse(raw)
                        if pre_data.laterality in ["R", "L"]:
                            laterality = pre_data.laterality
                    except Exception:
                        laterality = None
                    volume_string = "{}_{}_{}".format(
                        chunk.patient_db_id, chunk.study_id, chunk.series_id
                    )
                    if laterality and (volume_string not in laterality_dict):
                        laterality_dict[volume_string] = laterality

                elif chunk.type == 10019:  # contour data
                    raw = f.read(16)
                    contour_data = e2e_binary.contour_structure.parse(raw)

                    if contour_data.width > 0:
                        volume_string = "{}_{}_{}".format(
                            chunk.patient_db_id, chunk.study_id, chunk.series_id
                        )
                        slice_id = int(chunk.slice_id / 2)
                        contour_name = f"contour{contour_data.id}"
                        try:
                            raw_volume = np.frombuffer(
                                f.read(contour_data.width * 4), dtype=np.float32
                            )
                            contour = np.array(raw_volume)
                            max_float = np.finfo(np.float32).max
                            contour[(contour < 1e-9) | (contour == max_float)] = np.nan
                        except Exception as e:
                            warnings.warn(
                                (
                                    f"Could not read contour "
                                    f"image id {volume_string}"
                                    f"contour name {contour_name} "
                                    f"slice id {slice_id}."
                                ),
                                UserWarning,
                            )
                        else:
                            (
                                contour_dict[volume_string][contour_name][slice_id]
                            ) = contour

                elif chunk.type == 1073741824:  # image data
                    raw = f.read(20)
                    image_data = e2e_binary.image_structure.parse(raw)

                    if chunk.ind == 1:  # oct data
                        count = image_data.height * image_data.width
                        if count == 0:
                            break
                        raw_volume = np.fromfile(f, dtype=np.uint16, count=count)
                        volume_string = "{}_{}_{}".format(
                            chunk.patient_db_id, chunk.study_id, chunk.series_id
                        )
                        try:
                            image = LUT[raw_volume].reshape(
                                image_data.height, image_data.width
                            )
                        except Exception:
                            warnings.warn(
                                (
                                    f"Could not reshape image id {volume_string} with "
                                    f"{len(LUT[raw_volume])} elements into a "
                                    f"{image_data.height}x"
                                    f"{image_data.width} array"
                                ),
                                UserWarning,
                            )
                        else:
                            if legacy_intensity_transform:
                                image = pow(image, 1.0 / 2.4)
                            else:
                                image = self.vol_intensity_transform(image)

                            if volume_string in volume_array_dict.keys():
                                volume_array_dict[volume_string][
                                    int(chunk.slice_id / 2)
                                ] = image
                            else:
                                # try to capture these additional images
                                if volume_string in volume_array_dict_additional.keys():
                                    volume_array_dict_additional[volume_string].append(
                                        image
                                    )
                                else:
                                    volume_array_dict_additional[volume_string] = [
                                        image
                                    ]

            contour_data = {}
            for volume_id, contours in contour_dict.items():
                if volume_id in volume_dict:
                    num_slices = int(volume_dict[volume_id]) + 1
                else:
                    num_slices = None
                contour_data[volume_id] = {
                    k: [None] * (num_slices or len(v)) for k, v in contours.items()
                }

                for contour_name, contour_values in contours.items():
                    for slice_id, contour in contour_values.items():
                        (contour_data[volume_id][contour_name][slice_id]) = contour

            # Read metadata to attach to OCTVolumeWithMetaData
            metadata = self.read_all_metadata()

            oct_volumes = []
            for key, volume in chain(
                volume_array_dict.items(), volume_array_dict_additional.items()
            ):
                # remove any initalised volumes that never had image data attached
                volume = [slc for slc in volume if not isinstance(slc, int)]
                if volume is None or len(volume) == 0:
                    continue
                oct_volumes.append(
                    OCTVolumeWithMetaData(
                        volume=volume,
                        patient_id=self.patient_id,
                        first_name=self.first_name,
                        surname=self.surname,
                        sex=self.sex,
                        patient_dob=self.birthdate,
                        acquisition_date=self.acquisition_date,
                        volume_id=key,
                        laterality=laterality_dict.get(key),
                        contours=contour_data.get(key),
                        pixel_spacing=self.pixel_spacing,
                        metadata=metadata,
                    )
                )

        return oct_volumes

    def read_fundus_image(
        self,
        extract_scan_repeats: bool = False,
        scalex: float = 0.01,
    ) -> list[FundusImageWithMetaData]:
        """Reads fundus data.

        Args:
            extract_scan_repeats: if True, extract all fundus images, including those that appear repeated. Defaults to False.
            scalex: Manually set scale of x axis

        Returns:
            A sequence of FundusImageWithMetaData.
        """
        with open(self.filepath, "rb") as f:
            # traverse in second pass and  get all subdirectories
            chunk_stack = []
            for position in self.directory_stack:
                f.seek(position + self.byte_skip)
                raw = f.read(52)
                directory_chunk = e2e_binary.main_directory_structure.parse(raw)

                for ii in range(directory_chunk.num_entries):
                    raw = f.read(44)
                    chunk = e2e_binary.sub_directory_structure.parse(raw)
                    if chunk.start > chunk.pos:
                        chunk_stack.append([chunk.start, chunk.size])

            # initalise dict to hold all the image volumes
            image_array_dict = {}
            laterality_dict = {}
            laterality = None

            # traverse all chunks and extract slices
            for start, pos in chunk_stack:
                f.seek(start + self.byte_skip)
                raw = f.read(60)
                try:
                    # Heidelberg's updated anonymization seems to cause problems with
                    # some chunks. Observed problems include an empty raw and problems
                    # with undecodable bytes. For now, these chunks are skipped...
                    chunk = e2e_binary.chunk_structure.parse(raw)
                except Exception:
                    continue

                if chunk.type == 9:  # patient data
                    raw = f.read(127)
                    try:
                        patient_data = e2e_binary.patient_id_structure.parse(raw)
                        self.sex = patient_data.sex
                        self.first_name = patient_data.first_name
                        self.surname = patient_data.surname
                        julian_birthdate = (patient_data.birthdate / 64) - 14558805
                        self.birthdate = self.julian_to_ymd(julian_birthdate)
                        self.patient_id = patient_data.patient_id
                    except Exception:
                        pass

                elif chunk.type == 3:  # scan preamble data
                    raw = f.read(chunk.size)
                    try:
                        pre_data = e2e_binary.pre_data.parse(raw)
                        if pre_data.laterality in ["R", "L"]:
                            laterality = pre_data.laterality
                    except Exception:
                        laterality = None
                    volume_string = "{}_{}_{}".format(
                        chunk.patient_db_id, chunk.study_id, chunk.series_id
                    )
                    if laterality and (volume_string not in laterality_dict):
                        laterality_dict[volume_string] = laterality

                elif chunk.type == 1073741824:  # image data
                    raw = f.read(20)
                    image_data = e2e_binary.image_structure.parse(raw)
                    count = image_data.height * image_data.width
                    if count == 0:
                        break
                    if chunk.ind == 0:  # fundus data
                        raw_volume = np.frombuffer(f.read(count), dtype=np.uint8)
                        image = np.array(raw_volume).reshape(
                            image_data.height, image_data.width
                        )

                        image_string = "{}_{}_{}".format(
                            chunk.patient_db_id, chunk.study_id, chunk.series_id
                        )
                        if (
                            image_string in image_array_dict.keys()
                            and extract_scan_repeats
                        ):
                            is_in_keys = True
                            while is_in_keys:
                                image_string = image_string + "_"
                                if image_string not in image_array_dict.keys():
                                    is_in_keys = False

                        image_array_dict[image_string] = image

            # Read metadata to attach to FundusImageWithMetaData
            metadata = self.read_all_metadata()

            fundus_images = []
            for key, image in image_array_dict.items():
                fundus_images.append(
                    FundusImageWithMetaData(
                        image=image,
                        patient_id=self.patient_id,
                        image_id=key,
                        laterality=laterality_dict[key]
                        if key in laterality_dict.keys()
                        else None,
                        metadata=metadata,
                        pixel_spacing=[scalex, scalex],
                    )
                )

        return fundus_images

    def read_all_metadata(self):
        """
        Reads all available metadata and returns a dictionary.

        The metadata is a raw dump of everything available.

        Returns:
            dictionary with all metadata.
        """

        def _convert_to_dict(container):
            """Converts a container object to a dictionary"""
            return dict(
                (name, getattr(container, name))
                for name in container
                if not name.startswith("_")
            )

        metadata = dict()
        metadata["image_data"] = []
        metadata["bscan_data"] = []
        metadata["patient_data"] = []
        metadata["laterality_data"] = []
        metadata["contour_data"] = []
        metadata["fundus_data"] = []
        metadata["device_data"] = []
        metadata["examined_structure"] = {}
        metadata["scan_pattern"] = {}
        metadata["enface_modality"] = {}
        metadata["oct_modality"] = {}
        metadata["localizer"] = []
        metadata["eye_data"] = []
        metadata["uid_data"] = []
        metadata["time_data"] = []
        metadata["additional_device_data"] = []

        with open(self.filepath, "rb") as f:
            # get all subdirectories
            chunk_stack = []

            for position in self.directory_stack:
                f.seek(position + self.byte_skip)
                raw = f.read(52)
                directory_chunk = e2e_binary.main_directory_structure.parse(raw)

                for ii in range(directory_chunk.num_entries):
                    raw = f.read(44)
                    chunk = e2e_binary.sub_directory_structure.parse(raw)
                    if chunk.start > chunk.pos:
                        chunk_stack.append([chunk.start, chunk.size])

            # traverse all chunks and extract slices
            for start, pos in chunk_stack:
                f.seek(start + self.byte_skip)
                raw = f.read(60)
                try:
                    # Heidelberg's updated anonymization seems to cause problems with
                    # some chunks. Observed problems include an empty raw and problems
                    # with undecodable bytes. For now, these chunks are skipped...
                    chunk = e2e_binary.chunk_structure.parse(raw)
                except Exception:
                    continue

                image_string = "{}_{}_{}".format(
                    chunk.patient_db_id, chunk.study_id, chunk.series_id
                )

                if chunk.type == 9:  # patient data
                    raw = f.read(127)
                    try:
                        patient_data = e2e_binary.patient_id_structure.parse(raw)
                        metadata["patient_data"].append(_convert_to_dict(patient_data))
                    except Exception:
                        pass

                elif chunk.type == 10004:  # bscan metadata
                    raw = f.read(104)
                    bscan_metadata = e2e_binary.bscan_metadata.parse(raw)
                    metadata["bscan_data"].append(_convert_to_dict(bscan_metadata))

                elif chunk.type == 1073741824:  # fundus data
                    raw = f.read(20)
                    fundus_data = e2e_binary.image_structure.parse(raw)
                    metadata["fundus_data"].append(_convert_to_dict(fundus_data))

                elif chunk.type == 11:  # laterality data
                    raw = f.read(20)
                    laterality_data = e2e_binary.lat_structure.parse(raw)
                    metadata["laterality_data"].append(
                        _convert_to_dict(laterality_data)
                    )

                elif chunk.type == 10019:  # contour data
                    raw = f.read(16)
                    contour_data = e2e_binary.contour_structure.parse(raw)
                    metadata["contour_data"].append(_convert_to_dict(contour_data))

                elif chunk.type == 1073741824:  # image data
                    raw = f.read(20)
                    image_data = e2e_binary.image_structure.parse(raw)
                    metadata["image_data"].append(_convert_to_dict(image_data))

                elif chunk.type == 9001:  # device data ("Heidelberg Retina Angiograph")
                    raw = f.read(chunk.size)
                    device_data = e2e_binary.device_name.parse(raw)
                    metadata["device_data"].append(_convert_to_dict(device_data))

                elif chunk.type == 9005:  # examined structure ("Retina")
                    raw = f.read(chunk.size)
                    structure_data = e2e_binary.examined_structure.parse(raw)
                    if image_string not in metadata["examined_structure"]:
                        metadata["examined_structure"][
                            image_string
                        ] = structure_data.text[0]

                elif chunk.type == 9006:  # scan pattern
                    raw = f.read(chunk.size)
                    scan_pattern = e2e_binary.scan_pattern.parse(raw)
                    if image_string not in metadata["scan_pattern"]:
                        metadata["scan_pattern"][image_string] = scan_pattern.text[0]

                elif chunk.type == 9007:  # enface_modality (i.e. IR, FA, ICGA)
                    raw = f.read(chunk.size)
                    enface = e2e_binary.enface_modality.parse(raw)
                    if image_string not in metadata["enface_modality"]:
                        metadata["enface_modality"][image_string] = enface.text[1]

                elif chunk.type == 9008:
                    raw = f.read(chunk.size)
                    oct_modality = e2e_binary.oct_modality.parse(raw)
                    if image_string not in metadata["oct_modality"]:
                        metadata["oct_modality"][image_string] = oct_modality.text[0]

                elif chunk.type == 10025:
                    raw = f.read(chunk.size)
                    localizer = e2e_binary.localizer.parse(raw)
                    metadata["localizer"].append(_convert_to_dict(localizer))

                elif chunk.type == 7:  # eye data
                    raw = f.read(chunk.size)
                    eye_data = e2e_binary.eye_data.parse(raw)
                    metadata["eye_data"].append(_convert_to_dict(eye_data))

                elif chunk.type == 39:  # time zone, possibly timestamps
                    raw = f.read(chunk.size)
                    time_data = e2e_binary.time_data.parse(raw)
                    metadata["time_data"].append(_convert_to_dict(time_data))

                elif chunk.type in [52, 54, 1000, 1001]:  # various UIDs
                    try:
                        raw = f.read(chunk.size)
                        uid_data = e2e_binary.uid_data.parse(raw)
                        metadata["uid_data"].append(
                            {chunk.type: _convert_to_dict(uid_data)}
                        )
                    except StreamError:
                        pass

                # Chunks 1005, 1006, and 1007 seem to contain strings of device data,
                # including some servicers and distributors and other entities,
                # but not always in the same order.
                elif chunk.type in [1005, 1006]:
                    raw = f.read(chunk.size)
                    metadata["additional_device_data"].append(
                        {chunk.type: raw.decode()}
                    )

                elif chunk.type == 1007:
                    raw = f.read(chunk.size)
                    unknown_data = e2e_binary.unknown_data.parse(raw)
                    metadata["additional_device_data"].append(
                        {chunk.type: _convert_to_dict(unknown_data)}
                    )

        return metadata

    def read_custom_float(self, bytes: str) -> float:
        """Implementation of bespoke float type used in .e2e files.

        Notes:
            Custom float is a floating point type with no sign, 6-bit exponent, and 10-bit mantissa.

        Args:
            bytes: the two bytes.

        Returns:
            float
        """
        # convert two bytes to 16-bit binary representation
        bits = bin(bytes[0])[2:].zfill(8)[::-1] + bin(bytes[1])[2:].zfill(8)[::-1]

        # get mantissa and exponent
        mantissa = bits[:10]
        exponent = bits[10:]

        # convert to decimal representations
        mantissa_sum = 1 + int(mantissa, 2) / self.power
        exponent_sum = int(exponent[::-1], 2) - 63
        decimal_value = mantissa_sum * pow(2, exponent_sum)
        return decimal_value

    def uint16_to_ufloat16(self, uint16: int) -> float:
        """Implementation of bespoke float type used in .e2e files.

        Notes:
            Custom float is a floating point type with no sign, 6-bit exponent, and 10-bit mantissa.

        Args:
            uint16

        Returns:
            float
        """
        bits = "{0:016b}".format(uint16)[::-1]
        # get mantissa and exponent
        mantissa = bits[:10]
        exponent = bits[10:]
        exponent = exponent[::-1]

        # convert to decimal representations
        mantissa_sum = 1 + int(mantissa, 2) / self.power
        exponent_sum = int(exponent, 2) - 63
        decimal_value = mantissa_sum * np.float_power(2, exponent_sum)
        return decimal_value

    def vol_intensity_transform(self, data: np.array) -> np.array:
        """Implementation of intensity transform used in .e2e files.

        Notes:
            Code thanks to @oli4, see discussion in https://github.com/marksgraham/OCT-Converter/issues/21#issuecomment-1057455183
        """
        selection_0 = data == np.finfo(np.float32).max
        selection_data = data <= 1

        new = np.log(data[selection_data] + 2.44e-04)
        new = (new + 8.3) / 8.285

        data[selection_data] = new
        data[selection_0] = 0
        data = np.clip(data, 0, 1)
        return data

    def julian_to_ymd(self, J):
        """Converts Julian Day to Gregorian YMD.

        see https://en.wikipedia.org/wiki/Julian_day
        with thanks to https://github.com/seanredmond/juliandate
        """
        y = 4716
        j = 1401
        m = 2
        n = 12
        r = 4
        p = 1461
        v = 3
        u = 5
        s = 153
        w = 2
        B = 274277
        C = -38

        f = J + j + int(((int((4 * J + B) / 146097)) * 3) / 4) + C
        e = r * f + v
        g = int((e % p) / r)
        h = u * g + w

        D = int((h % s) / u) + 1
        M = ((int(h / s) + m) % n) + 1
        Y = int(e / p) - y + int((n + m - M) / n)

        return date(Y, M, D)



import numpy as np
from construct import (
    PaddedString,
    Int16un,
    Struct,
    Int8un,
    Int32sn,
    Int32un,
    Int64un,
    Array,
    Float32l,
)
from oct_converter.image_types import OCTVolumeWithMetaData, FundusImageWithMetaData
#import julian
import datetime
import collections


class E2E(object):
    """Class for extracting data from Heidelberg's .e2e file format.

    Notes:
        Mostly based on description of .e2e file format here:
        https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format.

    Attributes:
        filepath (str): Path to .img file for reading.
        header_structure (obj:Struct): Defines structure of volume's header.
        main_directory_structure (obj:Struct): Defines structure of volume's main directory.
        sub_directory_structure (obj:Struct): Defines structure of each sub directory in the volume.
        chunk_structure (obj:Struct): Defines structure of each data chunk.
        image_structure (obj:Struct): Defines structure of image header.
    """

    e2e_metadata_dict = {}

    def __init__(self, filepath):
        self.filepath = filepath
        self.header_structure = Struct(
            "magic" / PaddedString(12, "ascii"),
            "version" / Int32un,
            "unknown" / Array(10, Int16un),
        )
        self.main_directory_structure = Struct(
            "magic" / PaddedString(12, "ascii"),
            "version" / Int32un,
            "unknown" / Array(10, Int16un),
            "num_entries" / Int32un,
            "current" / Int32un,
            "prev" / Int32un,
            "unknown3" / Int32un,
        )
        self.sub_directory_structure = Struct(
            "pos" / Int32un,
            "start" / Int32un,
            "size" / Int32un,
            "unknown" / Int32un,
            "patient_id" / Int32un,
            "study_id" / Int32un,
            "series_id" / Int32un,
            "slice_id" / Int32sn,
            "unknown2" / Int16un,
            "unknown3" / Int16un,
            "type" / Int32un,
            "unknown4" / Int32un,
        )
        self.chunk_structure = Struct(
            "magic" / PaddedString(12, "ascii"),
            "unknown" / Int32un,
            "unknown2" / Int32un,
            "pos" / Int32un,
            "size" / Int32un,
            "unknown3" / Int32un,
            "patient_id" / Int32un,
            "study_id" / Int32un,
            "series_id" / Int32un,
            "slice_id" / Int32sn,
            "ind" / Int16un,
            "unknown4" / Int16un,
            "type" / Int32un,
            "unknown5" / Int32un,
        )
        self.image_structure = Struct(
            "size" / Int32un,
            "type" / Int32un,
            "unknown" / Int32un,
            "width" / Int32un,
            "height" / Int32un,
        )
        """
        self.patient_struct = Struct(
            "name" / PaddedString(31, "ascii"),
            "surname" / PaddedString(66, "ascii"),
            "dob" / Int32un,
            "sex" / PaddedString(1, "ascii"),
            "patient_id" / Int32un,
        )
        """
        self.patient_struct = Struct(
            "first_name" / PaddedString(31, "ascii"),
            "surname" / PaddedString(51, "ascii"),
            "title" / PaddedString(15, "ascii"),
            "birthdate" / Int32un,
            "sex" / PaddedString(1, "ascii"),
            "patient_id" / PaddedString(25, "ascii"),
        )
            
        self.slodataelement_struct = Struct(
            "datetime" / Int64un,
            "transform" / Array(6, Float32l),
        )
        self.lat_structure = Struct(
            "unknown" / Array(14, Int8un),
            "laterality" / Int8un,
            "unknown2" / Int8un,
        )
        self.bscan_metadata_struct = Struct(
            "unknown1" / Int32un,
            "imgSizeX" / Int32un,
            "imgSizeY" / Int32un,
            "posX1" / Float32l,
            "posY1" / Float32l,
            "posX2" / Float32l,
            "posY2" / Float32l,
            "zero1" / Int32un,
            "unknown2" / Float32l,
            "scaleY" / Float32l,
            "unknown3" / Float32l,
            "zero2" / Int32un,
            "unknown4" / Array(2, Float32l),
            "zero3" / Int32un,
            "imgSizeWidth" / Int32un,
            "numImages" / Int32un,
            "aktImage" / Int32un,
            "scantype" / Int32un,
            "centerPosX" / Float32l,
            "centerPosY" / Float32l,
            "unknown5" / Int32un,
            "acquisitionTime" / Int64un,
            "unknown6" / Int32un,
            "numAve" / Int32un,
            "unknown7" / Int32un,
            "imageQuality" / Float32l,
        )

    def write_deidentified(self, outfile_path):
        # read file to get all patient_struct locations
        self.read_oct_volume()

        if len(self.patient_struct_locs) != 1:
            print(f"ERROR WITH FILE {self.filepath}")
            print(f"patient struct locations: {self.patient_struct_locs}")
            return

        with open(self.filepath, "rb") as f:
            with open(outfile_path, "wb") as out_f:
                # iterate all bscan meta data location
                end = 0
                for bscan_struct_loc in sorted(self.bscan_struct_locs):
                    out_f.write(f.read(bscan_struct_loc - end))
                    raw = f.read(self.bscan_metadata_struct.sizeof())
                    bscan_data = self.bscan_metadata_struct.parse(raw)
                    
                    # use the original value except acquisitionTime
                    unknown1 = self.bscan_metadata_struct._subcons["unknown1"].build(bscan_data.unknown1)
                    imgSizeX = self.bscan_metadata_struct._subcons["imgSizeX"].build(bscan_data.imgSizeX)
                    imgSizeY = self.bscan_metadata_struct._subcons["imgSizeY"].build(bscan_data.imgSizeY)
                    posX1 = self.bscan_metadata_struct._subcons["posX1"].build(bscan_data.posX1)
                    posY1 = self.bscan_metadata_struct._subcons["posY1"].build(bscan_data.posY1)
                    posX2 = self.bscan_metadata_struct._subcons["posX2"].build(bscan_data.posX2)
                    posY2 = self.bscan_metadata_struct._subcons["posY2"].build(bscan_data.posY2)
                    zero1 = self.bscan_metadata_struct._subcons["zero1"].build(bscan_data.zero1)
                    unknown2 = self.bscan_metadata_struct._subcons["unknown2"].build(bscan_data.unknown2)
                    scaleY = self.bscan_metadata_struct._subcons["scaleY"].build(bscan_data.scaleY)
                    unknown3 = self.bscan_metadata_struct._subcons["unknown3"].build(bscan_data.unknown3)
                    zero2 = self.bscan_metadata_struct._subcons["zero2"].build(bscan_data.zero2)
                    unknown4 = self.bscan_metadata_struct._subcons["unknown4"].build(bscan_data.unknown4)
                    zero3 = self.bscan_metadata_struct._subcons["zero3"].build(bscan_data.zero3)
                    imgSizeWidth = self.bscan_metadata_struct._subcons["imgSizeWidth"].build(bscan_data.imgSizeWidth)
                    numImages = self.bscan_metadata_struct._subcons["numImages"].build(bscan_data.numImages)
                    aktImage = self.bscan_metadata_struct._subcons["aktImage"].build(bscan_data.aktImage)
                    scantype = self.bscan_metadata_struct._subcons["scantype"].build(bscan_data.scantype)
                    centerPosX = self.bscan_metadata_struct._subcons["centerPosX"].build(bscan_data.centerPosX)
                    centerPosY = self.bscan_metadata_struct._subcons["centerPosY"].build(bscan_data.centerPosY)
                    unknown5 = self.bscan_metadata_struct._subcons["unknown5"].build(bscan_data.unknown5)
                    acquisitionTime = self.bscan_metadata_struct._subcons["acquisitionTime"].build(116444720880000000)
                    unknown6 = self.bscan_metadata_struct._subcons["unknown6"].build(bscan_data.unknown6)
                    numAve = self.bscan_metadata_struct._subcons["numAve"].build(bscan_data.numAve)
                    unknown7 = self.bscan_metadata_struct._subcons["unknown7"].build(bscan_data.unknown7)
                    imageQuality = self.bscan_metadata_struct._subcons["imageQuality"].build(bscan_data.imageQuality)

                    items = [unknown1, imgSizeX, imgSizeY, posX1, posY1, posX2, posY2, zero1, unknown2, scaleY, unknown3,
                            zero2, unknown4, zero3, imgSizeWidth, numImages, aktImage, scantype, centerPosX, centerPosY,
                            unknown5, acquisitionTime, unknown6, numAve, unknown7, imageQuality]
                    for item in items:
                        out_f.write(item)
                        
                    end = bscan_struct_loc + self.bscan_metadata_struct.sizeof() 

                # end of for loop
                
                patient_struct_loc = self.patient_struct_locs[0]
                if len(self.bscan_struct_locs)==0:
                    return "Can not find B-Scan..."
                if patient_struct_loc < sorted(self.bscan_struct_locs)[-1]:
                    return "Patient Info comes earlier than B-Scan..."
                
                # write up until patient struct
                out_f.write(f.read(patient_struct_loc - end))
                # f.seek(patient_struct_loc)

                raw = f.read(self.patient_struct.sizeof())
                pt_data = self.patient_struct.parse(raw)

                first_name = self.patient_struct._subcons["first_name"].build("FIRSTNAME")
                last_name = self.patient_struct._subcons["surname"].build("LASTNAME")
                title = self.patient_struct._subcons["title"].build("TITLE")
                dob = self.patient_struct._subcons["birthdate"].build(19000101)
                sex = self.patient_struct._subcons["sex"].build("N")
                patient_id = self.patient_struct._subcons["patient_id"].build("PATIENT_ID")
                

                out_f.write(first_name)
                out_f.write(last_name)
                out_f.write(title)
                out_f.write(dob)
                out_f.write(sex)
                out_f.write(patient_id)
                # write rest of file
                out_f.write(f.read())
                return 0 

    def get_metadata(self, f):
        MAGIC_OFFSET = 116444720880000000  # 369 years
        raw = f.read(self.bscan_metadata_struct.sizeof())
        bscan_metadata = self.bscan_metadata_struct.parse(raw)
        timestamp = bscan_metadata.acquisitionTime
        timestamp_sec = int((timestamp - MAGIC_OFFSET) / 10000000)

        converted_date = datetime.datetime.fromtimestamp(timestamp_sec)

        return {
            "timestamp": converted_date,
            "image_index": bscan_metadata.aktImage,
            "num_slices": bscan_metadata.numImages,
            "scantype": bscan_metadata.scantype,
            "width": bscan_metadata.imgSizeX,
            "height": bscan_metadata.imgSizeY,
            "scaleY": bscan_metadata.scaleY,
            "x1": bscan_metadata.posX1,
            "y1": bscan_metadata.posY1,
            "x2": bscan_metadata.posX2,
            "y2": bscan_metadata.posY2,
        }

    def read_oct_volume(self, uint8_out=True):
        """Reads OCT data.

        Returns:
            obj:OCTVolumeWithMetaData
        """
        DATE_INT_OFFSET = 14555840.5

        most_recent_bscan_metadata_dict = {}
        oct_volume_dict = {}
        patient_metadata_dict = {}
        header_offset = 0
        self.patient_struct_locs = []
        self.bscan_struct_locs = []
        self.laterality = None

        with open(self.filepath, "rb") as f:
            try:
                raw = f.read(self.header_structure.sizeof())
                header = self.header_structure.parse(raw)

                raw = f.read(self.main_directory_structure.sizeof())
                main_directory = self.main_directory_structure.parse(raw)
            except UnicodeDecodeError:
                header_offset = 0x40
                f.seek(header_offset)
                raw = f.read(self.header_structure.sizeof())
                header = self.header_structure.parse(raw)

                raw = f.read(self.main_directory_structure.sizeof())
                main_directory = self.main_directory_structure.parse(raw)

            # traverse list of main directories in first pass
            directory_stack = []

            current = main_directory.current
            while current != 0:
                directory_stack.append(current)
                f.seek(current + header_offset)
                raw = f.read(self.main_directory_structure.sizeof())
                directory_chunk = self.main_directory_structure.parse(raw)
                current = directory_chunk.prev

            # traverse in second pass and  get all subdirectories
            chunk_stack = []
            volume_dict = {}
            for position in directory_stack:
                f.seek(position + header_offset)
                raw = f.read(self.main_directory_structure.sizeof())
                directory_chunk = self.main_directory_structure.parse(raw)

                for ii in range(directory_chunk.num_entries):
                    raw = f.read(self.sub_directory_structure.sizeof())
                    chunk = self.sub_directory_structure.parse(raw)

                    if chunk.type == 0x40000000:
                        volume_string = "{}_{}_{}".format(
                            chunk.patient_id, chunk.study_id, chunk.series_id
                        )
                        if volume_string not in volume_dict.keys():
                            volume_dict[volume_string] = chunk.slice_id / 2
                        elif chunk.slice_id / 2 > volume_dict[volume_string]:
                            volume_dict[volume_string] = chunk.slice_id / 2

                    if chunk.start > chunk.pos:
                        chunk_stack.append([chunk.start, chunk.size])

            # initalise dict to hold all the image volumes
            volume_array_dict = {}
            for volume, num_slices in volume_dict.items():
                if num_slices > 0:
                    volume_array_dict[volume] = [0] * int(num_slices)

            for start, pos in chunk_stack:
                f.seek(start + header_offset)
                raw = f.read(self.chunk_structure.sizeof())
                try:
                    chunk = self.chunk_structure.parse(raw)
                except Exception as e:
                    continue

                if chunk.type == 0x2714:  # bscan metadata
                    most_recent_bscan_metadata_dict = self.get_metadata(f)
                    bscan_struct_loc = start + 60 + header_offset
                    self.bscan_struct_locs.append(bscan_struct_loc)

                if chunk.type == 0x00000009:
                    patient_struct_loc = start + 60 + header_offset
                    self.patient_struct_locs.append(patient_struct_loc)
                    f.seek(patient_struct_loc)
                    raw = f.read(self.patient_struct.sizeof())
                    pt_data = self.patient_struct.parse(raw)
                    #try:
                    #    dob = julian.from_jd((pt_data.dob / 64.0) - DATE_INT_OFFSET)
                    #except Exception as e:
                    dob = None
                    patient_metadata_dict.update(
                        {
                            "dob": dob,
                            "name": pt_data.first_name,
                            "surname": pt_data.surname,
                            "sex": pt_data.sex,
                        }
                    )

                if chunk.type == 0xB:  # laterality data
                    raw = f.read(self.lat_structure.sizeof())
                    try:
                        laterality_data = self.lat_structure.parse(raw)
                        if laterality_data.laterality == 82:
                            self.laterality = "R"
                        elif laterality_data.laterality == 76:
                            self.laterality = "L"
                    except:
                        self.laterality = None

                if chunk.type == 0x40000000:  # image data
                    raw = f.read(self.image_structure.sizeof())
                    image_data = self.image_structure.parse(raw)

                    volume_string = "{}_{}_{}".format(
                        chunk.patient_id, chunk.study_id, chunk.series_id
                    )
                    slice_index = int(chunk.slice_id / 2) - 1

                    if chunk.ind == 1:  # oct data
                        # read entire image chunk
                        byte_chunk = f.read(2 * image_data.height * image_data.width)

                        # read bytes
                        image_byte_array = np.frombuffer(byte_chunk, dtype=np.uint16)

                        # reverse bits and repack
                        image_byte_array = np.packbits(
                            np.unpackbits(image_byte_array.view(np.uint8))[::-1]
                        ).view(np.uint16)[::-1]

                        # grab bottom 6 bits and reverse order
                        exponent_array = image_byte_array & 0x3F
                        # exponent_array = np.packbits(np.unpackbits(exponent_array.view(np.uint8))[::-1]).view(np.uint16)[::-1] >> 10
                        exponent_array = (
                            np.packbits(
                                np.unpackbits(exponent_array.view(np.uint8))[::-1]
                            ).view(np.uint16)[::-1]
                            >> 10
                        )
                        # manipulate exponent
                        exponent_array = exponent_array.view(np.int16) - 63

                        # grab the top 10 bits & manipulate mantissa
                        mantissa_array = 1 + (image_byte_array >> 6) / 1024

                        unshaped_image = mantissa_array * np.float_power(
                            2, exponent_array
                        )
                        # image = 256 * pow(image, 1.0 / 2.4)
                        unshaped_image = np.float_power(unshaped_image, 1.0 / 3.8)

                        # reshape image into width and height size
                        try:
                            image = unshaped_image.reshape(
                                image_data.width, image_data.height
                            )
                        except Exception as e:
                            break

                        if uint8_out:
                            # convert to uint8 explicitly to avoid annoying errors
                            image = (255 * (image / image.max())).astype(np.uint8)

                        if volume_string not in oct_volume_dict:
                            oct_volume_dict[volume_string] = {}

                        if "image_index" in most_recent_bscan_metadata_dict:
                            oct_volume_dict[volume_string][
                                most_recent_bscan_metadata_dict["image_index"]
                            ] = {
                                "image": image,
                                "laterality": self.laterality,
                                "metadata": most_recent_bscan_metadata_dict,
                            }
                        else:
                            print(
                                f"missing image index from {most_recent_bscan_metadata_dict}"
                            )

                        if volume_string in volume_array_dict.keys():
                            volume_array_dict[volume_string][slice_index] = image
                        else:
                            print(
                                "Failed to save image data for volume {}".format(
                                    volume_string
                                )
                            )

            oct_volumes = []
            for key, volume in volume_array_dict.items():
                if len(volume) > 1:
                    # only use OCT volumes with more than 1 slice, the 1 slice volumes
                    # appear to be fundus photos
                    oct_volume = OCTVolumeWithMetaData(volume=volume, patient_id=key)
                    oct_volumes.append(oct_volume)
        return oct_volumes, oct_volume_dict, patient_metadata_dict

    def read_fundus_image(self):
        """Reads fundus data.
        Returns:
            obj:FundusImageWithMetaData
        """
        with open(self.filepath, "rb") as f:
            metadata_dict = {}
            raw = f.read(36)
            header = self.header_structure.parse(raw)

            raw = f.read(52)
            main_directory = self.main_directory_structure.parse(raw)

            # traverse list of main directories in first pass
            directory_stack = []
            laterality = None
            current = main_directory.current
            while current != 0:
                directory_stack.append(current)
                f.seek(current)
                raw = f.read(52)
                directory_chunk = self.main_directory_structure.parse(raw)
                current = directory_chunk.prev

            # traverse in second pass and  get all subdirectories
            chunk_stack = []
            for position in directory_stack:
                f.seek(position)
                raw = f.read(52)
                directory_chunk = self.main_directory_structure.parse(raw)

                for ii in range(directory_chunk.num_entries):
                    raw = f.read(44)
                    chunk = self.sub_directory_structure.parse(raw)
                    if chunk.start > chunk.pos:
                        chunk_stack.append([chunk.start, chunk.size])

            # initalise dict to hold all the image volumes
            image_array_dict = {}
            laterality_dict = {}
            # traverse all chunks and extract slices
            for start, pos in chunk_stack:
                f.seek(start)
                raw = f.read(60)
                try:
                    chunk = self.chunk_structure.parse(raw)
                except Exception as e:
                    continue

                if chunk.type == 11:  # laterality data
                    raw = f.read(20)
                    try:
                        laterality_data = self.lat_structure.parse(raw)
                        if laterality_data.laterality == 82:
                            laterality = "R"
                        elif laterality_data.laterality == 76:
                            laterality = "L"
                    except Exception:
                        laterality = None

                if chunk.type == 0x2714:  # bscan metadata
                    metadata_dict = self.get_metadata(f)

                if chunk.type == 0x40000000:  # image data
                    raw = f.read(self.image_structure.sizeof())
                    image_data = self.image_structure.parse(raw)
                    if chunk.ind == 0:  # fundus data
                        raw_volume = np.fromstring(
                            f.read(image_data.height * image_data.width), dtype=np.uint8
                        )
                        image = np.array(raw_volume).reshape(
                            image_data.height, image_data.width
                        )
                        image_string = "{}_{}_{}".format(
                            chunk.patient_id, chunk.study_id, chunk.series_id
                        )
                        image_array_dict[image_string] = image
                        laterality_dict[image_string] = laterality

            fundus_images = []
            for key, image in image_array_dict.items():
                fundus_images.append(
                    FundusImageWithMetaData(
                        image=image,
                        patient_id=key,
                        laterality=laterality_dict[key]
                        if key in laterality_dict.keys()
                        else None,
                    )
                )

        return fundus_images, metadata_dict



from construct import (
    Array,
    Float32l,
    Float64l,
    Int8un,
    Int16un,
    Int32sn,
    Int32un,
    Int64un,
    PaddedString,
    Struct,
    this,
)

# Mostly based on description of .e2e file format here:
#         https://bitbucket.org/uocte/uocte/wiki/Heidelberg%20File%20Format.

header_structure = Struct(
    "magic1" / PaddedString(12, "ascii"),
    "version" / Int32un,
    "unknown" / Array(10, Int16un),
)
main_directory_structure = Struct(
    "magic2" / PaddedString(12, "ascii"),
    "version" / Int32un,
    "unknown" / Array(10, Int16un),
    "num_entries" / Int32un,
    "current" / Int32un,
    "prev" / Int32un,
    "unknown3" / Int32un,
)
sub_directory_structure = Struct(
    "pos" / Int32un,
    "start" / Int32un,
    "size" / Int32un,
    "unknown" / Int32un,
    # Patient DB ID is set by the software
    # and is not necessarily equal to patient_id_structure's patient_id
    "patient_db_id" / Int32un,
    "study_id" / Int32un,
    "series_id" / Int32un,
    "slice_id" / Int32sn,
    "unknown2" / Int16un,
    "unknown3" / Int16un,
    "type" / Int32un,
    "unknown4" / Int32un,
)
chunk_structure = Struct(
    "magic3" / PaddedString(12, "ascii"),
    "unknown" / Int32un,
    "unknown2" / Int32un,
    "pos" / Int32un,
    "size" / Int32un,
    "unknown3" / Int32un,
    # Patient DB ID is set by the software
    # and is not necessarily equal to patient_id_structure's patient_id
    "patient_db_id" / Int32un,
    "study_id" / Int32un,
    "series_id" / Int32un,
    "slice_id" / Int32sn,
    "ind" / Int16un,
    "unknown4" / Int16un,
    "type" / Int32un,
    "unknown5" / Int32un,
)
image_structure = Struct(
    "size" / Int32un,
    "type" / Int32un,
    "unknown" / Int32un,
    "height" / Int32un,
    "width" / Int32un,
)
patient_id_structure = Struct(
    "first_name" / PaddedString(31, "ascii"),
    "surname" / PaddedString(51, "ascii"),
    "title" / PaddedString(15, "ascii"),
    "birthdate" / Int32un,
    "sex" / PaddedString(1, "ascii"),
    "patient_id" / PaddedString(25, "ascii"),
)
lat_structure = Struct(
    "unknown" / Array(14, Int8un),
    "laterality" / PaddedString(1, "ascii"),
    "unknown2" / Int8un,
)
contour_structure = Struct(
    "unknown0" / Int32un,
    "id" / Int32un,
    "unknown1" / Int32un,
    "width" / Int32un,
)

# following the spec from
# https://github.com/neurodial/LibE2E/blob/d26d2d9db64c5f765c0241ecc22177bb0c440c87/E2E/dataelements/bscanmetadataelement.cpp#L75
bscan_metadata = Struct(
    "unknown1" / Int32un,
    "imgSizeY" / Int32un,
    "imgSizeX" / Int32un,
    "posX1" / Float32l,
    "posY1" / Float32l,
    "posX2" / Float32l,
    "posY2" / Float32l,
    "zero1" / Int32un,
    "unknown2" / Float32l,
    "scaley" / Float32l,
    "unknown3" / Float32l,
    "zero2" / Int32un,
    "unknown4" / Array(2, Float32l),
    "zero3" / Int32un,
    "imgSizeWidth" / Int32un,
    "numImages" / Int32un,
    "aktImage" / Int32un,
    "scanType" / Int32un,
    "centrePosX" / Float32l,
    "centrePosY" / Float32l,
    "unknown5" / Int32un,
    "acquisitionTime" / Int64un,
    "numAve" / Int32un,
    "imgQuality" / Float32l,
)

# Chunk 7: Eye Data (libE2E)
eye_data = Struct(
    "eyeSide" / PaddedString(1, "ascii"),
    "iop_mmHg" / Float64l,
    "refraction_dpt" / Float64l,
    "c_curve_mm" / Float64l,
    "vfieldMean" / Float64l,
    "vfieldVar" / Float64l,
    "cylinder_dpt" / Float64l,
    "axis_deg" / Float64l,
    "correctiveLens" / Int16un,
    "pupilSize_mm" / Float64l,
)

# 9001 Device Name
# Files examined have n_strings=3, string_size=256,
# text=["Heidelberg Retina Angiograph", "HRA", ""]
device_name = Struct(
    "n_strings" / Int32un,
    "string_size" / Int32un,
    "text" / Array(this.n_strings, PaddedString(this.string_size, "u16")),
)

# 9005 Examined Structure
# Files examined have n_strings=1, string_size=256,
# text=["Retina"]
examined_structure = Struct(
    "n_strings" / Int32un,
    "string_size" / Int32un,
    "text" / Array(this.n_strings, PaddedString(this.string_size, "u16")),
)

# 9006 Scan Pattern
# Files examined have n_strings=2, string_size=256,
# and scan patterns including "OCT Art Volume", "Images", "OCT B-SCAN",
# "3D Volume", "OCT Star Scan"
scan_pattern = Struct(
    "n_strings" / Int32un,
    "string_size" / Int32un,
    "text" / Array(this.n_strings, PaddedString(this.string_size, "u16")),
)

# 9007 Enface Modality
# Files examined have n_strings=2, string_size=256,
# and modalities including ["Infra-Red", "IR"],
# ["Fluroescein Angiography", "FA"], ["ICG Angiography", "ICGA"]
enface_modality = Struct(
    "n_strings" / Int32un,
    "string_size" / Int32un,
    "text" / Array(this.n_strings, PaddedString(this.string_size, "u16")),
)

# 9008 OCT Modality
# Files examined have n_strings=2, string_size=256, text=["OCT", "OCT"]
oct_modality = Struct(
    "n_strings" / Int32un,
    "string_size" / Int32un,
    "text" / Array(this.n_strings, PaddedString(this.string_size, "u16")),
)

# 10025 Localizer
# From eyepy; "transform" is described as "Parameters of affine transformation"
localizer = Struct(
    "unknown" / Array(6, Float32l),
    "windate" / Int32un,
    "transform" / Array(6, Float32l),
)

# 3 seems to indicate the start of the chunk pattern
# Examined files seem to have a mostly-regular pattern of 3, 2, ..., 5, 39
# Both chunks 3 and 5 seem to include laterality info
pre_data = Struct(
    "unknown" / Int32un,
    "laterality" / PaddedString(1, "ascii"),
    # There's more here that I'm unsure of.
    # There seems to be an "ART" in this chunk.
)

# 39 has some time zone data
time_data = Struct(
    "unknown" / Array(46, Int32un),
    "timezone1" / PaddedString(66, "u16"),
    "unknown2" / Array(9, Int16un),
    "timezone2" / PaddedString(66, "u16"),
    # There's more in this chunk (possibly datetimes, given tz)
    # and the chunk size varies.
)

# 52, 54, 1000, 1001 seem to be UIDs with padded strings
# 1000 may be StudyInstanceUID
uid_data = Struct("uid" / PaddedString(64, "ascii"))

# 1007 padded string with a brand name
unknown_data = Struct("unknown" / PaddedString(64, "ascii"))
